;;; custom-load.el --- automatically extracted custom dependencies

;;; Code:

(autoload 'custom-add-loads "cus-load")

(custom-add-loads 'ecb-general '("ecb-mode-line" "ecb-tod" "ecb"))
(custom-add-loads 'ecb-create-layout '("ecb-create-layout"))
(custom-add-loads 'ecb-layout '("ecb-compilation" "ecb-create-layout" "ecb-layout"))
(custom-add-loads 'ecb-face-options '("ecb-face"))
(custom-add-loads 'tools '("ecb"))
(custom-add-loads 'ecb-download '("ecb-upgrade"))
(custom-add-loads 'ecb-history '("ecb-face" "ecb"))
(custom-add-loads 'ecb '("ecb-compilation" "ecb-cycle" "ecb-eshell" "ecb-face" "ecb-help" "ecb-layout" "ecb-upgrade" "ecb"))
(custom-add-loads 'ecb-methods '("ecb-face" "ecb"))
(custom-add-loads 'ecb-directories '("ecb-face" "ecb-layout" "ecb"))
(custom-add-loads 'ecb-sources '("ecb-face" "ecb"))
(custom-add-loads 'ecb-eshell '("ecb-eshell"))
(custom-add-loads 'faces '("ecb-face"))
(custom-add-loads 'ecb-help '("ecb-help"))
(custom-add-loads 'ecb-non-semantic '("ecb"))
(custom-add-loads 'ecb-compilation-content '("ecb-compilation"))

;;; custom-load.el ends here
