;;; DO NOT MODIFY THIS FILE
(if (featurep 'ecb-autoloads) (error "Already loaded"))

;;;### (autoloads nil "_pkg" "ecb/_pkg.el")

(package-provide 'ecb :version 1.08 :type 'regular)

;;;***

;;;### (autoloads (ecb-show-help) "ecb-help" "ecb/ecb-help.el")

(autoload 'ecb-show-help "ecb-help" "\
Shows the online help of ECB in Info or HTML-format depending on the value
of the option `ecb-show-help-format'. If called with prefix argument, i.e. if
FORMAT is not nil then the user is prompted to choose the format of the help
\(Info or Html).

If an error about not finding the needed help-file occurs please take a look
at the options `ecb-help-info-start-file' and `ecb-help-html-start-file'!

Note: If you got ECB as a standard XEmacs-package maybe the
HTML-online-documentation is not included." t nil)

;;;***

;;;### (autoloads (ecb-byte-compile ecb-minor-mode ecb-activate) "ecb" "ecb/ecb.el")

(autoload 'ecb-activate "ecb" "\
Activates the ECB and creates all the buffers and draws the ECB-screen
with the actually choosen layout (see `ecb-layout-name'). This function raises
always the ECB-frame if called from another frame." t nil)

(autoload 'ecb-minor-mode "ecb" "\
Toggle ECB minor mode.
With prefix argument ARG, turn on if positive, otherwise off. Return non-nil
if the minor mode is enabled.

\\{ecb-mode-map}" t nil)

(autoload 'ecb-byte-compile "ecb" "\
Bytecompiles the ECB package. This is done for all lisp-files of ECB if
FORCE-ALL is not nil or for each lisp-file FILE.el which is either newer than
FILE.elc or if FILE.elc doesn't exist." t nil)

;;;***

(provide 'ecb-autoloads)
